/* With an argument of seconds, counts down when it receives SIGUSR1 until
 * SIGUSR2. If you send SIGUSR1 again it will continue counting down. Returns 0
 * when the count reaches zero or EXIT_FAILURE otherwise.
 *
 * Will print seconds remaining to stdout and exit with EXIT_FAILURE on SIGHUP.
 * 
 * Written by Tim Sutherland (timsuth@ihug.co.nz) and released into the public
 * domain on Mon Feb 28 20:59:38 NZDT 2000.
 */

#include <time.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

int sigsetup(void);
void sighandler(int sig);
int sighandler_write(int c);
unsigned int timeout(int fd, unsigned int secs);
void helpinfo(char *progname);

static volatile sig_atomic_t sighandler_pipe;

int main(int argc, char *argv[])
{
	unsigned int secs;
	int fd;

	if (argc!=2) {
		helpinfo(argv[0]);
		return EXIT_FAILURE;
	}
	secs = strtoul(argv[1], NULL, 10);
	if (secs>=UINT_MAX) {
		/* alarm uses unsigned int, we use (unsigned int)-1 for error */
		fprintf(stderr, "%s: seconds argument is too large\n", argv[0]);
		return EXIT_FAILURE;
	}
	fd = sigsetup();
	if (-1==fd)
		return EXIT_FAILURE;
	while (secs) {
		secs = timeout(fd, secs);
		if (secs==(unsigned int)-1)
			return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}

int sigsetup(void)
{
	struct sigaction cursig;
	int filedes[2];

	memset(&cursig, 0, sizeof(cursig));
	cursig.sa_handler = sighandler;
	sigaddset(&cursig.sa_mask, SIGUSR1);
	sigaddset(&cursig.sa_mask, SIGUSR2);
	sigaddset(&cursig.sa_mask, SIGALRM);
	sigaddset(&cursig.sa_mask, SIGHUP);
	if (-1==sigaction(SIGUSR1, &cursig, NULL) ||
			-1==sigaction(SIGUSR2, &cursig, NULL) ||
			-1==sigaction(SIGALRM, &cursig, NULL) ||
			-1==sigaction(SIGHUP, &cursig, NULL))
		return -1;
	if (-1==pipe(filedes))
		return -1;
	sighandler_pipe = filedes[1];
	return filedes[0];
}

void sighandler(int sig)
{
	switch(sig) {
	case SIGUSR1: sighandler_write('c'); /* CONNECT */
		      break;
	case SIGUSR2: sighandler_write('d'); /* DISCONNECT */
		      break;
	case SIGALRM: sighandler_write('t'); /* time is up */
		      break;
	case SIGHUP:  sighandler_write('h'); /* dump time remaining to stdout */
		      break;
	}
	return;
}

int sighandler_write(int c)
{
	return write(sighandler_pipe, &c, sizeof(c));
}

unsigned int timeout(int fd, unsigned int secs)
{
	unsigned int tmpsecs;
	int ret;
	int c;

	do {
		ret = read(fd, &c, sizeof(c));
	} while (ret==-1 && errno==EINTR);
	if (ret==-1) {
		perror("read failed");
		return -1;
	}
	switch(c) {
	case 'c': alarm(secs);
		  return secs;
	case 'd': tmpsecs = alarm(0);
		  if (tmpsecs)
			  secs = tmpsecs;
		  return secs;
	case 't': return 0;
	case 'h': tmpsecs = alarm(0);
		  if (tmpsecs)
			  secs = tmpsecs;
		  printf("%d\n", secs);
		  return -1;
	}
	fprintf(stderr, "invalid message from signal handler\n");
	return -1;
}

void helpinfo(char *progname)
{
	fprintf(stderr, "Usage: %s seconds\n", progname);
	return;
}
